from binance_f.impl import RestApiRequest
from binance_f.impl.utils.apisignature import create_signature
from binance_f.impl.utils.inputchecker import *
from binance_f.impl.utils.timeservice import *
from binance_f.impl.utils.urlparamsbuilder import UrlParamsBuilder
from binance_f.model import *


# For develop
# from binance_f.base.printobject import *


class RestApiRequestImpl(object):

    def __init__(self, api_key, secret_key, server_url="https://fapi.binance.com"):
        self.__api_key = api_key
        self.__secret_key = secret_key
        self.__server_url = server_url

    def __create_request_by_get(self, url, builder):
        request = RestApiRequest()
        request.method = "GET"
        request.host = self.__server_url
        request.header.update({'Content-Type': 'application/json'})
        request.url = url + "?" + builder.build_url()
        return request

    def __create_request_by_get_with_apikey(self, url, builder):
        request = RestApiRequest()
        request.method = "GET"
        request.host = self.__server_url
        request.header.update({'Content-Type': 'application/json'})
        request.header.update({"X-MBX-APIKEY": self.__api_key})
        request.url = url + "?" + builder.build_url()
        #  # For develop
        # print("====== Request ======")
        # print(request)
        # PrintMix.print_data(request)
        # print("=====================")
        return request

    def __create_request_by_post_with_signature(self, url, builder):
        request = RestApiRequest()
        request.method = "POST"
        request.host = self.__server_url
        builder.put_url("recvWindow", 60000)
        builder.put_url("timestamp", str(get_current_timestamp() - 1000))
        create_signature(self.__secret_key, builder)
        request.header.update({'Content-Type': 'application/json'})
        request.header.update({"X-MBX-APIKEY": self.__api_key})
        request.post_body = builder.post_map
        request.url = url + "?" + builder.build_url()
        # # For develop
        # print("====== Request ======")
        # print(request)
        # PrintMix.print_data(request)
        # print("=====================")
        return request

    def __create_request_by_delete_with_signature(self, url, builder):
        request = RestApiRequest()
        request.method = "DELETE"
        request.host = self.__server_url
        builder.put_url("recvWindow", 60000)
        builder.put_url("timestamp", str(get_current_timestamp() - 1000))
        create_signature(self.__secret_key, builder)
        request.header.update({'Content-Type': 'application/json'})
        request.header.update({"X-MBX-APIKEY": self.__api_key})
        request.url = url + "?" + builder.build_url()
        # # For develop
        # print("====== Request ======")
        # print(request)
        # PrintMix.print_data(request)
        # print("=====================")
        return request

    def __create_request_by_get_with_signature(self, url, builder):
        request = RestApiRequest()
        request.method = "GET"
        request.host = self.__server_url
        builder.put_url("recvWindow", 60000)
        builder.put_url("timestamp", str(get_current_timestamp() - 1000))
        create_signature(self.__secret_key, builder)
        request.header.update({"Content-Type": "application/x-www-form-urlencoded"})
        request.header.update({"X-MBX-APIKEY": self.__api_key})
        request.url = url + "?" + builder.build_url()
        # # For develop
        # print("====== Request ======")
        # print(request)
        # PrintMix.print_data(request)
        # print("=====================")
        return request

    def __create_request_by_put_with_signature(self, url, builder):
        request = RestApiRequest()
        request.method = "PUT"
        request.host = self.__server_url
        builder.put_url("recvWindow", 60000)
        builder.put_url("timestamp", str(get_current_timestamp() - 1000))
        create_signature(self.__secret_key, builder)
        request.header.update({'Content-Type': 'application/json'})
        request.header.update({"X-MBX-APIKEY": self.__api_key})
        request.url = url + "?" + builder.build_url()
        # # For develop
        # print("====== Request ======")
        # print(request)
        # PrintMix.print_data(request)
        # print("=====================")
        return request

    def get_servertime(self):
        builder = UrlParamsBuilder()
        request = self.__create_request_by_get("/fapi/v1/time", builder)

        def parse(json_wrapper):
            result = json_wrapper.get_int("serverTime")
            return result

        request.json_parser = parse
        return request

    def get_exchange_information(self):
        builder = UrlParamsBuilder()
        request = self.__create_request_by_get("/fapi/v1/exchangeInfo", builder)

        def parse(json_wrapper):
            result = ExchangeInformation.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def get_order_book(self, symbol, limit):
        check_should_not_none(symbol, "symbol")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get("/fapi/v1/depth", builder)

        def parse(json_wrapper):
            result = OrderBook.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def get_recent_trades_list(self, symbol, limit):
        check_should_not_none(symbol, "symbol")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get("/fapi/v1/trades", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = Trade.json_parse(item)
                result.append(element.toDict())
            return result

        request.json_parser = parse
        return request

    def get_old_trade_lookup(self, symbol, limit, fromId):
        check_should_not_none(symbol, "symbol")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("limit", limit)
        builder.put_url("fromId", fromId)

        request = self.__create_request_by_get_with_apikey("/fapi/v1/historicalTrades", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = Trade.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_aggregate_trades_list(self, symbol, fromId, startTime, endTime, limit):
        check_should_not_none(symbol, "symbol")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("fromId", fromId)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get("/fapi/v1/aggTrades", builder)

        def parse(json_wrapper):
            aggregate_trades_list = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                trade = AggregateTrade.json_parse(item)
                aggregate_trades_list.append(trade)
            return aggregate_trades_list

        request.json_parser = parse
        return request

    def get_candlestick_data(self, symbol, interval, startTime, endTime, limit):
        check_should_not_none(symbol, "symbol")
        check_should_not_none(symbol, "interval")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("interval", interval)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get("/fapi/v1/klines", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = Candlestick.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_mark_price(self, symbol):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)

        request = self.__create_request_by_get("/fapi/v1/premiumIndex", builder)

        def parse(json_wrapper):
            result = list()
            if symbol:
                result = MarkPrice.json_parse(json_wrapper)
            else:
                data_list = json_wrapper.convert_2_array()
                for item in data_list.get_items():
                    element = MarkPrice.json_parse(item)
                    result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_funding_rate(self, symbol, startTime, endTime, limit):
        check_should_not_none(symbol, "symbol")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get("/fapi/v1/fundingRate", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = FundingRate.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_ticker_price_change_statistics(self, symbol):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)

        request = self.__create_request_by_get("/fapi/v1/ticker/24hr", builder)

        def parse(json_wrapper):
            result = list()

            if symbol:
                element = TickerPriceChangeStatistics.json_parse(json_wrapper)
                result.append(element)
            else:
                data_list = json_wrapper.convert_2_array()
                for item in data_list.get_items():
                    element = TickerPriceChangeStatistics.json_parse(item)
                    result.append(element)

            return result

        request.json_parser = parse
        return request

    def get_symbol_price_ticker(self, symbol):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)

        request = self.__create_request_by_get("/fapi/v1/ticker/price", builder)

        def parse(json_wrapper):
            result = list()
            if symbol:
                element = SymbolPrice.json_parse(json_wrapper)
                result.append(element)
            else:
                data_list = json_wrapper.convert_2_array()
                for item in data_list.get_items():
                    element = SymbolPrice.json_parse(item)
                    result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_symbol_orderbook_ticker(self, symbol):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)

        request = self.__create_request_by_get("/fapi/v1/ticker/bookTicker", builder)

        def parse(json_wrapper):
            result = list()
            if symbol:
                element = SymbolOrderBook.json_parse(json_wrapper)
                result.append(element)
            else:
                data_list = json_wrapper.convert_2_array()
                for item in data_list.get_items():
                    element = SymbolOrderBook.json_parse(item)
                    result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_open_interest(self, symbol):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)

        request = self.__create_request_by_get("/fapi/v1/openInterest", builder)

        def parse(json_wrapper):
            result = list()
            element = OpenInterest.json_parse(json_wrapper)
            result.append(element)
            return element

        request.json_parser = parse
        return request

    def get_liquidation_orders(self, symbol, startTime, endTime, limit):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get("/fapi/v1/allForceOrders", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = LiquidationOrder.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def change_position_mode(self, dualSidePosition):
        check_should_not_none(dualSidePosition, "dualSidePosition")
        builder = UrlParamsBuilder()
        builder.put_url("dualSidePosition", dualSidePosition)

        request = self.__create_request_by_post_with_signature("/fapi/v1/positionSide/dual", builder)

        def parse(json_wrapper):
            result = CodeMsg.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def get_position_mode(self):

        request = self.__create_request_by_get_with_signature("/fapi/v1/positionSide/dual", builder)

        def parse(json_wrapper):
            result = PositionMode.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def post_order(self, symbol, side, ordertype, timeInForce, quantity, reduceOnly, price, newClientOrderId, stopPrice,
                   workingType, closePosition,
                   positionSide, callbackRate, activationPrice, priceProtect, newOrderRespType):
        check_should_not_none(symbol, "symbol")
        check_should_not_none(side, "side")
        check_should_not_none(ordertype, "ordertype")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("side", side)
        builder.put_url("type", ordertype)
        builder.put_url("timeInForce", timeInForce)
        builder.put_url("quantity", quantity)
        builder.put_url("reduceOnly", reduceOnly)
        builder.put_url("price", price)
        builder.put_url("newClientOrderId", newClientOrderId)
        builder.put_url("stopPrice", stopPrice)
        builder.put_url("workingType", workingType)
        builder.put_url("closePosition", closePosition)
        builder.put_url("positionSide", positionSide)
        builder.put_url("callbackRate", callbackRate)
        builder.put_url("activationPrice", activationPrice)
        builder.put_url("priceProtect", priceProtect)
        builder.put_url("newOrderRespType", newOrderRespType)

        request = self.__create_request_by_post_with_signature("/fapi/v1/order", builder)

        def parse(json_wrapper):
            result = Order.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def post_batch_order(self, batchOrders):
        check_should_not_none(batchOrders, "batchOrders")

        builder = UrlParamsBuilder()
        builder.put_url("batchOrders", batchOrders)

        request = self.__create_request_by_post_with_signature("/fapi/v1/batchOrders", builder)

        def parse(json_wrapper):
            result = list()
            dataList = json_wrapper.convert_2_array()
            for item in dataList.get_items():
                if item.contain_key("code"):
                    result.append(Msg.json_parse(item))
                else:
                    result.append(Order.json_parse(item))
            return result

        request.json_parser = parse
        return request

    def get_order(self, symbol, orderId, origClientOrderId):
        check_should_not_none(symbol, "symbol")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("orderId", orderId)
        builder.put_url("origClientOrderId", origClientOrderId)

        request = self.__create_request_by_get_with_signature("/fapi/v1/order", builder)

        def parse(json_wrapper):
            result = Order.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def cancel_order(self, symbol, orderId, origClientOrderId):
        check_should_not_none(symbol, "symbol")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("orderId", orderId)
        builder.put_url("origClientOrderId", origClientOrderId)

        request = self.__create_request_by_delete_with_signature("/fapi/v1/order", builder)

        def parse(json_wrapper):
            result = Order.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def cancel_all_orders(self, symbol):
        check_should_not_none(symbol, "symbol")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)

        request = self.__create_request_by_delete_with_signature("/fapi/v1/allOpenOrders", builder)

        def parse(json_wrapper):
            result = CodeMsg.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def cancel_list_orders(self, symbol, orderIdList, origClientOrderIdList):
        check_should_not_none(symbol, "symbol")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("orderIdList", orderIdList)
        builder.put_url("origClientOrderIdList", origClientOrderIdList)
        request = self.__create_request_by_delete_with_signature("/fapi/v1/batchOrders", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                if item.contain_key("code"):
                    element = Msg.json_parse(item)
                else:
                    element = Order.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_open_orders(self, symbol):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)

        request = self.__create_request_by_get_with_signature("/fapi/v1/openOrders", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = Order.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_all_orders(self, symbol, orderId, startTime, endTime, limit):
        check_should_not_none(symbol, "symbol")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("orderId", orderId)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get_with_signature("/fapi/v1/allOrders", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = Order.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_balance(self):
        builder = UrlParamsBuilder()

        request = self.__create_request_by_get_with_signature("/fapi/v1/balance", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = Balance.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_account_information(self):
        builder = UrlParamsBuilder()

        request = self.__create_request_by_get_with_signature("/fapi/v1/account", builder)

        def parse(json_wrapper):
            result = AccountInformation.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def change_initial_leverage(self, symbol, leverage):
        check_should_not_none(symbol, "symbol")
        check_should_not_none(leverage, "leverage")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("leverage", leverage)

        request = self.__create_request_by_post_with_signature("/fapi/v1/leverage", builder)

        def parse(json_wrapper):
            result = Leverage.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def change_margin_type(self, symbol, marginType):
        check_should_not_none(symbol, "symbol")
        check_should_not_none(marginType, "marginType")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("marginType", marginType)

        request = self.__create_request_by_post_with_signature("/fapi/v1/marginType", builder)

        def parse(json_wrapper):
            result = CodeMsg.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def change_position_margin(self, symbol, amount, type):
        check_should_not_none(symbol, "symbol")
        check_should_not_none(amount, "amount")
        check_should_not_none(type, "type")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("amount", amount)
        builder.put_url("type", type)

        request = self.__create_request_by_post_with_signature("/fapi/v1/positionMargin", builder)

        def parse(json_wrapper):
            result = PositionMargin.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def get_position_margin_change_history(self, symbol, type, startTime, endTime, limit):
        check_should_not_none(symbol, "symbol")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("type", type)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get_with_signature("/fapi/v1/positionMargin/history", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = PositionMarginHist.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_position(self):
        builder = UrlParamsBuilder()

        request = self.__create_request_by_get_with_signature("/fapi/v1/positionRisk", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = Position.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_account_trades(self, symbol, startTime, endTime, fromId, limit):
        check_should_not_none(symbol, "symbol")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("fromId", fromId)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get_with_signature("/fapi/v1/userTrades", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = MyTrade.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_income_history(self, symbol, incomeType, startTime, endTime, limit):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("incomeType", incomeType)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get_with_signature("/fapi/v1/income", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = Income.json_parse(item)
                result.append(element.toDict())
            return result

        request.json_parser = parse
        return request

    def start_user_data_stream(self):
        builder = UrlParamsBuilder()

        request = self.__create_request_by_post_with_signature("/fapi/v1/listenKey", builder)

        def parse(json_wrapper):
            result = json_wrapper.get_string("listenKey")
            return result

        request.json_parser = parse
        return request

    def keep_user_data_stream(self):
        builder = UrlParamsBuilder()

        request = self.__create_request_by_put_with_signature("/fapi/v1/listenKey", builder)

        def parse(json_wrapper):
            result = "OK"
            return result

        request.json_parser = parse
        return request

    def close_user_data_stream(self):
        builder = UrlParamsBuilder()

        request = self.__create_request_by_delete_with_signature("/fapi/v1/listenKey", builder)

        def parse(json_wrapper):
            result = "OK"
            return result

        request.json_parser = parse
        return request

    def get_open_interest_stats(self, symbol, period, startTime, endTime, limit):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("period", period)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get("/futures/data/openInterestHist", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = OpenInterestStats.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_top_long_short_accounts(self, symbol, period, startTime, endTime, limit):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("period", period)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get("/futures/data/topLongShortAccountRatio", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = LongShortRatio.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_top_long_short_positions(self, symbol, period, startTime, endTime, limit):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("period", period)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get("/futures/data/topLongShortPositionRatio", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = LongShortRatio.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_global_long_short_accounts(self, symbol, period, startTime, endTime, limit):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("period", period)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get("/futures/data/globalLongShortAccountRatio", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = LongShortRatio.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_taker_buy_sell_ratio(self, symbol, period, startTime, endTime, limit):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("period", period)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get("/futures/data/takerlongshortRatio", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = TakerBuySellRatio.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_blvt_nav_candlestick_data(self, symbol, interval, startTime, endTime, limit):
        check_should_not_none(symbol, "symbol")
        check_should_not_none(symbol, "interval")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("interval", interval)
        builder.put_url("startTime", startTime)
        builder.put_url("endTime", endTime)
        builder.put_url("limit", limit)

        request = self.__create_request_by_get("/fapi/v1/lvtKlines", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = BLVTNAVCandlestick.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_composite_index_info(self, symbol):
        check_should_not_none(symbol, "symbol")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)

        request = self.__create_request_by_get("/fapi/v1/indexInfo", builder)

        def parse(json_wrapper):
            result = list()
            element = IndexInfo.json_parse(json_wrapper)
            result.append(element)
            return element

        request.json_parser = parse
        return request

    def auto_cancel_all_orders(self, symbol, countdownTime):
        check_should_not_none(symbol, "symbol")
        check_should_not_none(symbol, "countdownTime")
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)
        builder.put_url("countdownTime", countdownTime)

        request = self.__create_request_by_post_with_signature("/fapi/v1/countdownCancelAll", builder)

        def parse(json_wrapper):
            result = CountdownCancelAll.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def get_balance_v2(self):
        builder = UrlParamsBuilder()

        request = self.__create_request_by_get_with_signature("/fapi/v2/balance", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = BalanceV2.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_account_information_v2(self):
        builder = UrlParamsBuilder()

        request = self.__create_request_by_get_with_signature("/fapi/v2/account", builder)

        def parse(json_wrapper):
            result = AccountInformationV2.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request

    def get_position_v2(self, symbol):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)

        request = self.__create_request_by_get_with_signature("/fapi/v2/positionRisk", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = Position.json_parse(item)
                result.append(element.toDict())
            return result

        request.json_parser = parse
        return request

    def get_leverage_bracket(self, symbol):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)

        request = self.__create_request_by_get_with_signature("/fapi/v1/leverageBracket", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = LeverageBracket.json_parse(item)
                result.append(element)

            return result

        request.json_parser = parse
        return request

    def get_adl_quantile(self, symbol):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)

        request = self.__create_request_by_get_with_signature("/fapi/v1/adlQuantile", builder)

        def parse(json_wrapper):
            result = list()
            data_list = json_wrapper.convert_2_array()
            for item in data_list.get_items():
                element = AdlQuantile.json_parse(item)
                result.append(element)
            return result

        request.json_parser = parse
        return request

    def get_api_trading_stats(self, symbol):
        builder = UrlParamsBuilder()
        builder.put_url("symbol", symbol)

        request = self.__create_request_by_get_with_signature("/fapi/v1/apiTradingStatus", builder)

        def parse(json_wrapper):
            result = ApiTradingStatus.json_parse(json_wrapper)
            return result

        request.json_parser = parse
        return request



class SymbolTickerEvent:

    def __init__(self):
        self.eventType = ""
        self.eventTime = 0
        self.symbol = ""
        self.priceChange = 0.0
        self.priceChangePercent = 0.0
        self.weightedAvgPrice = 0.0
        self.lastPrice = 0.0
        self.lastQty = 0.0
        self.open = 0.0
        self.high = 0.0
        self.low = 0.0
        self.totalTradedBaseAssetVolume = 0.0
        self.totalTradedQuoteAssetVolume = 0.0
        self.openTime = 0
        self.closeTime = 0
        self.firstId = None
        self.lastId = None
        self.count = 0

    @staticmethod
    def json_parse(json_wrapper):
        ticker_event = SymbolTickerEvent()
        ticker_event.eventType = json_wrapper.get_string("e")
        ticker_event.eventTime = json_wrapper.get_int("E")
        ticker_event.symbol = json_wrapper.get_string("s")
        ticker_event.priceChange = json_wrapper.get_float("p")
        ticker_event.priceChangePercent = json_wrapper.get_float("P")
        ticker_event.weightedAvgPrice = json_wrapper.get_float("w")
        ticker_event.lastPrice = json_wrapper.get_float("c")
        ticker_event.lastQty = json_wrapper.get_float("Q")
        ticker_event.open = json_wrapper.get_float("o")
        ticker_event.high = json_wrapper.get_float("h")
        ticker_event.low = json_wrapper.get_float("l")
        ticker_event.totalTradedBaseAssetVolume = json_wrapper.get_float("v")
        ticker_event.totalTradedQuoteAssetVolume = json_wrapper.get_float("q")
        ticker_event.openTime = json_wrapper.get_int("O")
        ticker_event.closeTime = json_wrapper.get_int("C")
        ticker_event.firstId = json_wrapper.get_int("F")
        ticker_event.lastId = json_wrapper.get_int("L")
        ticker_event.count = json_wrapper.get_int("n")
        return ticker_event


import shutil

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

try:
    shutil.rmtree("./build")
    shutil.rmtree("./dist")
except OSError as e:
    print("Error: %s - %s." % (e.filename, e.strerror))

setuptools.setup(
    name="TradeGate",  # This is the name of the package
    version="0.3.8.6",
    author="Rustin Soraki",  # Full name of the author
    description="A Trading Gateway",
    long_description=long_description,  # Long description read from the readme file
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(
        where="TradeGates"
    ),  # List of all python modules to be installed
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],  # Information to filter the project on PyPi website
    python_requires=">=3.6",  # Minimum version requirement of the package
    py_modules=["TradeGate"],  # Name of the python package
    package_dir={"": "TradeGates"},  # Directory of the source code of the package
    install_requires=[
        "binance-connector",
        "requests",
        "apscheduler",
        "websocket-client",
        "urllib3",
        "tzlocal<3.0",
        "pandas",
        "pybit",
    ],  # Install other dependencies if any
)

# python setup.py sdist bdist_wheel
# twine check dist/*
# twine upload dist/*



import threading
import logging
from apscheduler.schedulers.blocking import BlockingScheduler
from binance_f.impl.websocketconnection import ConnectionState
from binance_f.impl.utils.timeservice import get_current_timestamp


def watch_dog_job(*args):
    watch_dog_instance = args[0]
    for connection in watch_dog_instance.connection_list:
        if connection.state == ConnectionState.CONNECTED:
            if watch_dog_instance.is_auto_connect:
                ts = get_current_timestamp() - connection.last_receive_time
                if ts > watch_dog_instance.receive_limit_ms:
                    watch_dog_instance.logger.warning("[Sub][" + str(connection.id) + "] No response from server")
                    connection.re_connect_in_delay(watch_dog_instance.connection_delay_failure)
        elif connection.in_delay_connection():
            watch_dog_instance.logger.warning("[Sub] call re_connect")
            connection.re_connect()
            pass
        elif connection.state == ConnectionState.CLOSED_ON_ERROR:
            if watch_dog_instance.is_auto_connect:
                connection.re_connect_in_delay(watch_dog_instance.connection_delay_failure)
                pass


class WebSocketWatchDog(threading.Thread):
    mutex = threading.Lock()
    connection_list = list()

    def __init__(self, is_auto_connect=True, receive_limit_ms=60000, connection_delay_failure=15):
        threading.Thread.__init__(self)
        self.is_auto_connect = is_auto_connect
        self.receive_limit_ms = receive_limit_ms
        self.connection_delay_failure = connection_delay_failure
        self.logger = logging.getLogger("binance-client")
        self.scheduler = BlockingScheduler()
        self.scheduler.add_job(watch_dog_job, "interval", max_instances=10, seconds=1, args=[self])
        self.start()

    def run(self):
        self.scheduler.start()

    def on_connection_created(self, connection):
        self.mutex.acquire()
        self.connection_list.append(connection)
        self.mutex.release()

    def on_connection_closed(self, connection):
        self.mutex.acquire()
        self.connection_list.remove(connection)
        self.mutex.release()



from kucoin.base_request.base_request import KucoinBaseRestApi


class UserData(KucoinBaseRestApi):

    def get_actual_fee(self, symbols):
        """
        https://docs.kucoin.top/#actual-fee-rate-of-the-trading-pair
        :param symbols: symbols
        :type: str
        :return:
            {
                "code": "200000",
                "data": [
                    {
                        "symbol": "BTC-USDT",
                        "takerFeeRate": "0.001",
                        "makerFeeRate": "0.001"
                    },
                    {
                        "symbol": "KCS-USDT",
                        "takerFeeRate": "0.002",
                        "makerFeeRate": "0.0005"
                    }
                ]
            }
        """
        params = {
            "symbols": symbols
        }
        
        return self._request('GET', '/api/v1/trade-fees', params=params)

    def get_base_fee(self):
        """
        https://docs.kucoin.top/#basic-user-fee

        :return:
            {
                "code": "200000",
                "data": {
                    "takerFeeRate": "0.001",
                    "makerFeeRate": "0.001"
                }
            }
        """
        return self._request('GET', '/api/v1/base-fee')

    def get_sub_user(self):
        """
        https://docs.kucoin.com/#get-user-info-of-all-sub-accounts

        :return:
            [{
                "userId": "5cbd31ab9c93e9280cd36a0a",  //subUserId
                "subName": "kucoin1",
                "remarks": "kucoin1"
            },
            {
                "userId": "5cbd31b89c93e9280cd36a0d",
                "subName": "kucoin2",
                "remarks": "kucoin2"
            }
            ]
        """
        return self._request('GET', '/api/v1/sub/user')

    def create_account(self, account_type, currency):
        """
        https://docs.kucoin.com/#create-an-account
        :param account_type: Account type (Mandatory)
        :type: str
        :param currency: Currency (Mandatory)
        :type: str
        :return:
        {
            "id": "5bd6e9286d99522a52e458de"  //accountId
        }
        """
        params = {
            'type': account_type,
            'currency': currency
        }
        return self._request('POST', '/api/v1/accounts', params=params)

    def get_account_list(self, currency=None, account_type=None):
        """
        https://docs.kucoin.com/#list-accounts
        :param currency: Currency (Optional)
        :type: str
        :param account_type: Account type (Optional)
        :type: str
        :return:
        [{
            "id": "5bd6e9286d99522a52e458de",  //accountId
            "currency": "BTC",  //Currency
            "type": "main",     //Account type, including main and trade
            "balance": "237582.04299",  //Total assets of a currency
            "available": "237582.032",  //Available assets of a currency
            "holds": "0.01099". //Hold assets of a currency
        },
        {
            "id": "5bd6e9216d99522a52e458d6",
            "currency": "BTC",
            "type": "trade",
            "balance": "1234356",
            "available": "1234356",
            "holds": "0"
        }]
        """
        params = {}
        if currency:
            params['currency'] = currency
        if account_type:
            params['type'] = account_type
        return self._request('GET', '/api/v1/accounts', params=params)

    def get_account(self, accountId):
        """
        https://docs.kucoin.com/#get-an-account
        :param accountId: ID of the account (Mandatory)
        :type: str
        :return:
        {
            "currency": "KCS",  //Currency
            "balance": "1000000060.6299",  //Total assets of a currency
            "available": "1000000060.6299",  //Available assets of a currency
            "holds": "0". //Hold assets of a currency
        }
        """
        return self._request('GET', '/api/v1/accounts/{accountId}'.format(accountId=accountId))

    def get_account_ledger(self, **kwargs):
        """
        https://docs.kucoin.top/#get-account-ledgers
        :param kwargs: [optional] currency, direction, bizType, startAt, endAt, currentPage , pageSize
        :return:
        {
            "currentPage": 1,
            "pageSize": 10,
            "totalNum": 3,
            "totalPage": 1,
            "items": [
                {
                    "id": "5bc7f080b39c5c03486eef8c",//unique key
                    "currency": "KCS",  //Currency
                    "amount": "0.0998", //Change amount of the funds
                    "fee": "0",  //Deposit or withdrawal fee
                    "balance": "0",  //Total assets of a currency
                    "bizType": "Withdraw",  //business type
                    "direction": "in",     //side, in or out
                    "createdAt": 1540296039000,  //Creation time
                    "context": {          //Business core parameters

                        "orderId": "5bc7f080b39c5c03286eef8a",
                        "txId": "bf848bfb6736780b930e12c68721ea57f8b0484a4af3f30db75c93ecf16905c9"
                    }
                },
                {
                    "id": "5bc7f080b39c5c03486def8c",//unique key
                    "currency": "KCS",
                    "amount": "0.0998",
                    "fee": "0",
                    "balance": "0",
                    "bizType": "Deposit",
                    "direction": "in",
                    "createdAt": 1540296039000,
                    "context": {

                        "orderId": "5bc7f080b39c5c03286eef8a",
                        "txId": "bf848bfb6736780b930e12c68721ea57f8b0484a4af3f30db75c93ecf16905c9"
                    }
                },
                {
                    "id": "5bc7f080b39c5c03486def8a",//unique key
                    "currency": "KCS",
                    "amount": "0.0998",
                    "fee": "0",
                    "balance": "0",
                    "bizType": "trade exchange",
                    "direction": "in",
                    "createdAt": 1540296039000,
                    "context": {

                        "tradeId": "5bc7f080b3949c03286eef8a",
                        "orderId": "5bc7f080b39c5c03286eef8e",
                        "symbol": "BTC-USD"
                    }
                }
            ]
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)

        return self._request('GET', '/api/v1/accounts/ledgers', params=params)

    def get_account_hold(self, accountId, **kwargs):
        """
        https://docs.kucoin.com/#get-holds
        :param accountId: ID of the account. (Mandatory)
        :type: str
        :param kwargs: [optional] currentPage , pageSize
        :return:
        {
            "currentPage": 1,
            "pageSize": 10,
            "totalNum": 2,
            "totalPage": 1,
            "items": [
                {
                    "currency": "ETH",  //Currency
                    "holdAmount": "5083",  //Hold amount of a currency
                    "bizType": "Withdraw",     //business type
                    "orderId": "5bc7f080b39c5c03286eef8e", // ID of funds freezed order
                    "createdAt": 1545898567000, //Creation time
                    "updatedAt": 1545898567000。//update time
                },
                {
                    "currency": "ETH",
                    "holdAmount": "1452",
                    "bizType": "Withdraw",
                    "orderId": "5bc7f518b39c5c033818d62d",
                    "createdAt": 1545898567000,
                    "updatedAt": 1545898567000
                }
            ]
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)
        return self._request('GET', '/api/v1/accounts/{accountId}/holds'.format(accountId=accountId), params=params)

    def get_sub_account(self, subUserId):
        """
        https://docs.kucoin.com/#get-account-balance-of-a-sub-account
        :param subUserId: the user ID of a sub-account. (Mandatory)
        :type: str
        :return:
        {
            "subUserId": "5caefba7d9575a0688f83c45",
            "subName": "sdfgsdfgsfd",
            "mainAccounts": [{
                "currency": "BTC",
                "balance": "8",
                "available": "8",
            "holds": "0",
            "baseCurrency": "BTC",
            "baseCurrencyPrice": "1",
            "baseAmount": "1.1"
            }],
            "tradeAccounts": [{
                "currency": "BTC",
                "balance": "1000",
                "available": "1000",
            "holds": "0",
            "baseCurrency": "BTC",
            "baseCurrencyPrice": "1",
            "baseAmount": "1.1"

          }],
          "marginAccounts": [{
            "currency": "BTC",
            "balance": "1.1",
            "available": "1.1",
            "holds": "0",
            "baseCurrency": "BTC",
            "baseCurrencyPrice": "1",
            "baseAmount": "1.1"
          }]
        }
        """
        return self._request('GET', '/api/v1/sub-accounts/{subUserId}'.format(subUserId=subUserId))

    def get_sub_accounts(self):
        """
        https://docs.kucoin.com/#get-the-aggregated-balance-of-all-sub-accounts
        :return:
        [
          {
                "subUserId": "5caefba7d9575a0688f83c45",
                "subName": "kucoin1",
                "mainAccounts": [{
                    "currency": "BTC",
                    "balance": "6",
                    "available": "6",
              "holds": "0",
              "baseCurrency": "BTC",
              "baseCurrencyPrice": "1",
              "baseAmount": "1.1"

                }],
                "tradeAccounts": [{
                    "currency": "BTC",
                    "balance": "1000",
                    "available": "1000",
              "holds": "0",
              "baseCurrency": "BTC",
              "baseCurrencyPrice": "1",
              "baseAmount": "1.1"
            }],
            "marginAccounts": [{
                "currency": "BTC",
                "balance": "1.1",
                "available": "1.1",
                "holds": "0",
                "baseCurrency": "BTC",
                "baseCurrencyPrice": "1",
                "baseAmount": "1.1"
            }]
          }
        ]
        """
        return self._request('GET', '/api/v1/sub-accounts')

    def get_transferable(self, currency, account_type):
        """
        https://docs.kucoin.com/#get-the-transferable
        :param currency: currency (Mandatory)
        :type: str
        :param account_type: The account type (Mandatory)
        :type: str
        :return:
         {
            "currency": "KCS",
            "balance": "0",
            "available": "0",
            "holds": "0",
            "transferable": "0"
        }
        """
        params = {
            'currency': currency,
            'type': account_type
        }
        return self._request('GET', '/api/v1/accounts/transferable', params=params)

    def transfer_master_sub(self, currency, amount, direction, subUserId, clientOid='', accountType=None,
                            subAccountType=None):
        """
        https://docs.kucoin.com/#transfer-between-master-user-and-sub-user
        :param currency: currency (Mandatory)
        :type: str
        :param amount: Transfer amount, the amount is a positive integer multiple of the currency precision. (Mandatory)
        :type: str
        :param direction: OUT — the master user to sub user,IN — the sub user to the master user. (Mandatory)
        :type: str
        :param accountType: The account type of the master user (Optional)
        :type: str
        :param subAccountType: The account type of the sub user (Optional)
        :type: str
        :param subUserId: the user ID of a sub-account. (Mandatory)
        :type: str
        :param clientOid: Unique order id created by users to identify their orders, e.g. UUID. (Mandatory)
        :type: str
        :return:
        {
            "orderId": "5cbd870fd9575a18e4438b9a"
        }
        """
        params = {
            'currency': currency,
            'amount': amount,
            'direction': direction,
            'subUserId': subUserId
        }
        if accountType:
            params['accountType'] = accountType
        if subAccountType:
            params['subAccountType'] = subAccountType
        if not clientOid:
            clientOid = self.return_unique_id
        params['clientOid'] = clientOid
        return self._request('POST', '/api/v2/accounts/sub-transfer', params=params)

    def inner_transfer(self, currency, from_payer, to_payee, amount, clientOid=''):
        """
        https://docs.kucoin.com/#inner-transfer
        :param currency: currency (Mandatory)
        :type: str
        :param from_payer: Account type of payer (Mandatory)
        :type: str
        :param to_payee: Account type of payee (Mandatory)
        :type: str
        :param amount: Transfer amount, the amount is a positive integer multiple of the currency precision. (Mandatory)
        :type: str
        :param clientOid: Unique order id created by users to identify their orders, e.g. UUID. (Mandatory)
        :type: str
        :return:
        {
            "orderId": "5bd6e9286d99522a52e458de"
        }
        """
        params = {
            'currency': currency,
            'from': from_payer,
            'to': to_payee,
            'amount': amount
        }
        if not clientOid:
            clientOid = self.return_unique_id
        params['clientOid'] = clientOid
        return self._request('POST', '/api/v2/accounts/inner-transfer', params=params)

    def create_deposit_address(self, currency, chain=None):
        """
        https://docs.kucoin.com/#create-deposit-address
        :param currency: currency (Mandatory)
        :type: str
        :param chain: The chain name of currency, e.g. The available value for USDT are OMNI, ERC20, TRC20,default is
        OMNI. This only apply for multi-chain currency, and there is no need for single chain currency. (Optional)
        :type: str
        :return:
        {
            "address": "0x78d3ad1c0aa1bf068e19c94a2d7b16c9c0fcd8b1",
            "memo": "5c247c8a03aa677cea2a251d",   //tag
            "chain": "OMNI"
        }
        """
        params = {
            'currency': currency
        }
        if chain:
            params['chain'] = chain
        return self._request('POST', '/api/v1/deposit-addresses', params=params)

    def get_deposit_addressv2(self, currency, chain=None):
        """
        https://docs.kucoin.com/#get-deposit-addresses-v2
        :param currency: currency (Mandatory)
        :type: str
        :param chain: The chain name of currency, e.g. The available value for USDT are OMNI, ERC20, TRC20,default is
        OMNI. This only apply for multi-chain currency, and there is no need for single chain currency. (Optional)
        :type: str
        :return:
        [{
            "address": "0x78d3ad1c0aa1bf068e19c94a2d7b16c9c0fcd8b1",
            "memo": "5c247c8a03aa677cea2a251d",        //tag
            "chain": "OMNI",
            "contractAddress": ""
        }]
        """
        params = {
            'currency': currency
        }
        if chain:
            params['chain'] = chain
        return self._request('GET', '/api/v2/deposit-addresses', params=params)

    def get_deposit_address(self, currency, chain=None):
        """
        https://docs.kucoin.com/#get-deposit-address
        :param currency: currency (Mandatory)
        :type: str
        :param chain: The chain name of currency, e.g. The available value for USDT are OMNI, ERC20, TRC20,default is
        OMNI. This only apply for multi-chain currency, and there is no need for single chain currency. (Optional)
        :type: str
        :return:
        {
            "address": "0x78d3ad1c0aa1bf068e19c94a2d7b16c9c0fcd8b1",
            "memo": "5c247c8a03aa677cea2a251d",        //tag
            "chain": "OMNI"
        }
        """
        params = {
            'currency': currency
        }
        if chain:
            params['chain'] = chain
        return self._request('GET', '/api/v1/deposit-addresses', params=params)

    def get_deposit_list(self, **kwargs):
        """
        https://docs.kucoin.com/#get-deposit-list
        :param kwargs: [optional] currency, startAt, endAt, status, currentPage, pageSize
        :return:
        {
          "currentPage": 1,
          "pageSize": 5,
          "totalNum": 2,
          "totalPage": 1,
            "items": [{
                "address": "0x5f047b29041bcfdbf0e4478cdfa753a336ba6989",
                "memo": "5c247c8a03aa677cea2a251d",
                "amount": 1,
                "fee": 0.0001,
                "currency": "KCS",
                "isInner": false,
                "walletTxId": "5bbb57386d99522d9f954c5a@test004",
                "status": "SUCCESS",
                "remark": "test",
                "createdAt": 1544178843000,
                "updatedAt": 1544178891000
            }, {
                "address": "0x5f047b29041bcfdbf0e4478cdfa753a336ba6989",
                "memo": "5c247c8a03aa677cea2a251d",
                "amount": 1,
                "fee": 0.0001,
                "currency": "KCS",
                "isInner": false,
                "walletTxId": "5bbb57386d99522d9f954c5a@test003",
                "status": "SUCCESS",
                "remark": "test",
                "createdAt": 1544177654000,
                "updatedAt": 1544178733000
            }]
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)
        return self._request('GET', '/api/v1/deposits', params=params)

    def get_deposit_list_v1(self, **kwargs):
        """
        https://docs.kucoin.com/#get-v1-historical-deposits-list
        :param kwargs: [optional] currency, startAt, endAt, status, currentPage, pageSize
        :return:
        {
          "currentPage": 1,
          "pageSize": 5,
          "totalNum": 2,
          "totalPage": 1,
            "items": [{
                "address": "0x5f047b29041bcfdbf0e4478cdfa753a336ba6989",
                "memo": "5c247c8a03aa677cea2a251d",
                "amount": 1,
                "fee": 0.0001,
                "currency": "KCS",
                "isInner": false,
                "walletTxId": "5bbb57386d99522d9f954c5a@test004",
                "status": "SUCCESS",
                "remark": "test",
                "createdAt": 1544178843000,
                "updatedAt": 1544178891000
            }, {
                "address": "0x5f047b29041bcfdbf0e4478cdfa753a336ba6989",
                "memo": "5c247c8a03aa677cea2a251d",
                "amount": 1,
                "fee": 0.0001,
                "currency": "KCS",
                "isInner": false,
                "walletTxId": "5bbb57386d99522d9f954c5a@test003",
                "status": "SUCCESS",
                "remark": "test",
                "createdAt": 1544177654000,
                "updatedAt": 1544178733000
            }]
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)
        return self._request('GET', '/api/v1/hist-deposits', params=params)

    def get_withdrawal_list(self, **kwargs):
        """
        https://docs.kucoin.com/#get-withdrawals-list
        :param kwargs: [optional] currency, status, startAt, endAt, currentPage , pageSize
        :return:
        {
            "currentPage": 1,
            "pageSize": 10,
            "totalNum": 1,
            "totalPage": 1,
            "items": [{
                "id": "5c2dc64e03aa675aa263f1ac",
                "address": "0x5bedb060b8eb8d823e2414d82acce78d38be7fe9",
                "memo": "",
                "currency": "ETH",
                "amount": 1.0000000,
                "fee": 0.0100000,
                "walletTxId": "3e2414d82acce78d38be7fe9",
                "isInner": false,
                "status": "FAILURE",
                "remark": "test",
                "createdAt": 1546503758000,
                "updatedAt": 1546504603000
            }]
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)

        return self._request('GET', '/api/v1/withdrawals', params=params)

    def get_hist_withdrawal_list(self, **kwargs):
        """
        https://docs.kucoin.com/#get-v1-historical-withdrawals-list
        :param kwargs: [optional] currency, status, startAt, endAt, currentPage , pageSize
        :return:
        {
            "currentPage": 1,
            "pageSize": 1,
            "totalNum": 2,
            "totalPage": 2,
            "items": [{
                "currency": "BTC",
                "createAt": 1526723468,
                "amount": "0.534",
                "address": "33xW37ZSW4tQvg443Pc7NLCAs167Yc2XUV",
                "walletTxId": "aeacea864c020acf58e51606169240e96774838dcd4f7ce48acf38e3651323f4",
                "isInner": false,
                "status": "SUCCESS"
            }]
        }
        """
        params = {}
        if kwargs:
            params.update(kwargs)

        return self._request('GET', '/api/v1/hist-withdrawals', params=params)

    def get_withdrawal_quota(self, currency, chain=None):
        """
        https://docs.kucoin.com/#get-withdrawal-quotas
        :param currency:  currency (Mandatory)
        :type: str
        :param chain: The chain name of currency, e.g. The available value for USDT are OMNI, ERC20, TRC20, default is
        OMNI. This only apply for multi-chain currency, and there is no need for single chain currency. (Optional)
        :type: str
        :return:
        {
            "currency": "KCS",
            "limitBTCAmount": "2.0",
            "usedBTCAmount": "0",
            "limitAmount": "75.67567568",
            "remainAmount": "75.67567568",
            "availableAmount": "9697.41991348",
            "withdrawMinFee": "0.93000000",
            "innerWithdrawMinFee": "0.00000000",
            "withdrawMinSize": "1.4",
            "isWithdrawEnabled": true,
            "precision": 8,   //withdrawal precision
            "chain": "OMNI"
        }
        """
        params = {
            'currency': currency
        }
        if chain:
            params['chain'] = chain
        return self._request('GET', '/api/v1/withdrawals/quotas', params=params)

    def apply_withdrawal(self, currency, address, amount, **kwargs):
        """
        https://docs.kucoin.com/#apply-withdraw-2
        :param currency: Currency. (Mandatory)
        :type: str
        :param address: Withdrawal address (Mandatory)
        :type: str
        :param amount: Withdrawal amount, a positive number which is a multiple of the amount precision
            (fees excluded) (Mandatory)
        :type: float
        :param kwargs:  [Optional]  memo, isInner, remark, chain
        :return:
         {
           "withdrawalId": "5bffb63303aa675e8bbe18f9"
         }
        """
        params = {
            'currency': currency,
            'address': address,
            'amount': amount
        }
        if kwargs:
            params.update(kwargs)

        return self._request('POST', '/api/v1/withdrawals', params=params)

    def cancel_withdrawal(self, withdrawalId):
        """
        https://docs.kucoin.com/#cancel-withdrawal
        :param withdrawalId: Path parameter, a unique ID for a withdrawal order  (Mandatory)
        :type: str
        :return:
        """
        return self._request('DELETE', '/api/v1/withdrawals/{withdrawalId}'.format(withdrawalId=withdrawalId))





class LongShortRatio:

    def __init__(self):
        self.symbol = ""
        self.longShortRatio = 0.0  #  long/short account num ratio of top traders
        self.longAccount = 0.0  # long account num ratio of top traders
        self.shortAccount = 0.0  # short account num ratio of top traders 
        self.timestamp = 0
    
    @staticmethod
    def json_parse(json_data):
        result = LongShortRatio()
        result.symbol = json_data.get_string("symbol")
        result.longShortRatio = json_data.get_float("longShortRatio")
        result.longAccount = json_data.get_float("longAccount")
        result.shortAccount = json_data.get_float("shortAccount")
        result.timestamp = json_data.get_int("timestamp")

        return result



from kucoin.base_request.base_request import KucoinBaseRestApi


class GetToken(KucoinBaseRestApi):

    def get_ws_token(self, is_private=False):
        """
        https://docs.kucoin.com/#apply-connect-token
        :param is_private private or public
        :return:
        """
        uri = '/api/v1/bullet-public'
        if is_private:
            uri = '/api/v1/bullet-private'

        return self._request('POST', uri, auth=is_private)





import asyncio
import json
import time
import websockets
from random import random
from uuid import uuid4
import logging

logger = logging.getLogger(__name__)


class ConnectWebsocket:
    MAX_RECONNECTS = 5
    MAX_RECONNECT_SECONDS = 60

    def __init__(self, loop, client, callback, private=False):
        self._loop = loop
        self._client = client
        self._callback = callback
        self._reconnect_num = 0
        self._ws_details = None
        self._connect_id = None
        self._private = private
        self._last_ping = None
        self._socket = None
        self._topics = []
        asyncio.ensure_future(self.run_forever(), loop=self._loop)

    @property
    def topics(self):
        return self._topics

    async def _run(self, event: asyncio.Event):
        keep_alive = True
        self._last_ping = time.time()  # record last ping
        self._ws_details = None
        self._ws_details = self._client.get_ws_token(self._private)
        logger.debug(self._ws_details)

        async with websockets.connect(self.get_ws_endpoint(), ssl=self.get_ws_encryption()) as socket:
            self._socket = socket
            self._reconnect_num = 0

            if not event.is_set():
                await self.send_ping()
                event.set()

            while keep_alive:
                if time.time() - self._last_ping > self.get_ws_pingtimeout():
                    await self.send_ping()
                try:
                    _msg = await asyncio.wait_for(self._socket.recv(), timeout=self.get_ws_pingtimeout())
                except asyncio.TimeoutError:
                    await self.send_ping()
                except asyncio.CancelledError:
                    logger.exception('CancelledError')
                    await self._socket.ping()
                else:
                    try:
                        msg = json.loads(_msg)
                    except ValueError:
                        logger.warning(_msg)
                    else:
                        await self._callback(msg)

    def get_ws_endpoint(self):
        if not self._ws_details:
            raise Exception("Websocket details Error")
        ws_connect_id = str(uuid4()).replace('-', '')
        token = self._ws_details['token']
        endpoint = self._ws_details['instanceServers'][0]['endpoint']
        ws_endpoint = f"{endpoint}?token={token}&connectId={ws_connect_id}"
        return ws_endpoint

    def get_ws_encryption(self):
        if not self._ws_details:
            raise Exception("Websocket details Error")
        return self._ws_details['instanceServers'][0]['encrypt']

    def get_ws_pingtimeout(self):
        if not self._ws_details:
            raise Exception("Websocket details Error")
        _timeout = int(self._ws_details['instanceServers'][0]['pingTimeout'] / 1000) - 2
        return _timeout

    async def run_forever(self):
        while True:
            await self._reconnect()

    async def _reconnect(self):
        logger.info('Websocket start connect/reconnect')

        self._reconnect_num += 1
        reconnect_wait = self._get_reconnect_wait(self._reconnect_num)
        logger.info(f'asyncio sleep reconnect_wait={reconnect_wait} s reconnect_num={self._reconnect_num}')
        await asyncio.sleep(reconnect_wait)
        logger.info(f'asyncio sleep ok')
        event = asyncio.Event()

        tasks = {
            asyncio.ensure_future(self._recover_topic_req_msg(event), loop=self._loop): self._recover_topic_req_msg,
            asyncio.ensure_future(self._run(event), loop=self._loop): self._run
        }

        while set(tasks.keys()):
            finished, pending = await asyncio.wait(tasks.keys(), return_when=asyncio.FIRST_EXCEPTION)
            exception_occur = False
            for task in finished:
                if task.exception():
                    exception_occur = True
                    logger.warning("{} got an exception {}".format(task, task.exception()))
                    for pt in pending:
                        logger.warning(f'pending {pt}')
                        try:
                            pt.cancel()
                        except asyncio.CancelledError:
                            logger.exception('CancelledError ')
                        logger.warning('cancel ok.')

            if exception_occur:
                break

        logger.warning('_reconnect over.')

    async def _recover_topic_req_msg(self, event):
        logger.info(f'recover topic event {self.topics} waiting')
        await event.wait()
        logger.info(f'recover topic event {self.topics} done.')
        for topic in self.topics:
            await self.send_message({
                'type': 'subscribe',
                'topic': topic,
                'response': True
            })
            logger.info(f'{topic} OK')

    def _get_reconnect_wait(self, attempts):
        expo = 2 ** attempts
        return round(random() * min(self.MAX_RECONNECT_SECONDS, expo - 1) + 1)

    async def send_ping(self):
        msg = {
            'id': str(int(time.time() * 1000)),
            'type': 'ping'
        }
        await self._socket.send(json.dumps(msg))
        self._last_ping = time.time()

    async def send_message(self, msg, retry_count=0):
        if not self._socket:
            if retry_count < self.MAX_RECONNECTS:
                await asyncio.sleep(1)
                await self.send_message(msg, retry_count + 1)
        else:
            msg['id'] = str(int(time.time() * 1000))
            msg['privateChannel'] = self._private
            await self._socket.send(json.dumps(msg))



class PositionMargin:

    def __init__(self):
        self.code = 0
        self.msg = ""
        self.amount = 0.0
        self.type = 0

    @staticmethod
    def json_parse(json_data):
        result = PositionMargin()
        result.code = json_data.get_int("code")
        result.msg = json_data.get_string("msg")
        result.amout = json_data.get_float("amount")
        result.type = json_data.get_int("type")

        return result



import logging
import warnings
from datetime import datetime

from pybit.unified_trading import HTTP

from Exchanges.BaseExchange import BaseExchange
from Utils import DataHelpers, BybitHelpers


def is_spot_order_data_valid(order_data: DataHelpers.OrderData):
    check_spot_order_mandatory_fields(order_data)

    check_spot_order_type(order_data)

    check_spot_order_side(order_data)

    check_spot_order_time_in_force(order_data)

    check_spot_order_limit_price(order_data)


def check_spot_order_limit_price(order_data):
    if order_data.order_type in ["LIMIT", "LIMIT_MAKER"] and order_data.price is None:
        raise ValueError("Price must be specified for limit orders.")


def check_spot_order_time_in_force(order_data):
    if order_data.time_in_force not in BybitExchange.spot_time_in_forces:
        raise ValueError(
            "Time-in-force not correctly specified. Available time-in-force for spot market: {}".format(
                BybitExchange.spot_time_in_forces
            )
        )


def check_spot_order_mandatory_fields(order_data):
    if (
        order_data.symbol is None
        or order_data.quantity is None
        or order_data.side is None
        or order_data.order_type is None
    ):
        raise ValueError("Missing mandatory fields.")


def check_spot_order_side(order_data):
    if order_data.side not in ["BUY", "SELL"]:
        raise ValueError("Order side can only be 'BUY' or 'SELL'")


def check_spot_order_type(order_data):
    if order_data.order_type not in BybitExchange.spot_order_types:
        raise ValueError(
            "Order type not correctly specified. Available order types for spot market: {}".format(
                BybitExchange.spot_order_types
            )
        )


def is_futures_order_data_valid(order_data: DataHelpers.FuturesOrderData):
    check_futures_order_symbol(order_data)
    check_futures_order_quantity(order_data)
    check_futures_order_side(order_data)
    check_futures_order_type(order_data)
    check_futures_order_time_in_force(order_data)
    check_futures_order_close_position(order_data)
    check_futures_order_reduce_only(order_data)

    if order_data.orderType.startswith("STOP"):
        check_futures_stop_order_datas(order_data)

    if "LIMIT" in order_data.orderType and order_data.price is None:
        raise ValueError("Specify 'price' for limit orders.")

    if order_data.extra_params is not None:
        check_futures_order_trigger_by(order_data)
        check_futures_order_tp_trigger_by(order_data)
        check_futures_order_sl_trigger_by(order_data)
        check_futures_order_position_idx(order_data)


def check_futures_order_position_idx(order_data):
    if "positionIdx" in order_data.extra_params.keys() and order_data.extra_params[
        "positionIdx"
    ] not in [0, 1, 2]:
        raise ValueError("'positionIdx' was not correctly specified.")


def check_futures_order_sl_trigger_by(order_data):
    if "slTriggerBy" in order_data.extra_params.keys() and order_data.extra_params[
        "slTriggerBy"
    ] not in ["LastPrice", "IndexPrice", "MarkPrice"]:
        raise ValueError("'slTriggerBy' was not correctly specified.")


def check_futures_order_tp_trigger_by(order_data):
    if "tpTriggerBy" in order_data.extra_params.keys() and order_data.extra_params[
        "tpTriggerBy"
    ] not in ["LastPrice", "IndexPrice", "MarkPrice"]:
        raise ValueError("'tpTriggerBy' was not correctly specified.")


def check_futures_order_trigger_by(order_data):
    if "triggerBy" in order_data.extra_params.keys() and order_data.extra_params[
        "tpTriggerBy"
    ] not in ["LastPrice", "IndexPrice", "MarkPrice"]:
        raise ValueError("'triggerBy' was not correctly specified.")


def check_futures_stop_order_datas(order_data):
    if order_data.extra_params is None:
        raise ValueError("Specify 'basePrice' in 'extraParams'")
    if "basePrice" not in order_data.extra_params.keys():
        raise ValueError("Specify 'basePrice' in 'extraParams'")
    if order_data.stop_price is None:
        raise ValueError("Specify 'stopPrice'.")


def check_futures_order_reduce_only(order_data):
    if order_data.reduce_only is None:
        raise ValueError("Specify reduceOnly.")


def check_futures_order_close_position(order_data):
    if order_data.close_position is None:
        raise ValueError("Specify closePosition.")


def check_futures_order_time_in_force(order_data):
    if order_data.time_in_force is None:
        raise ValueError("Specify timeInForce.")

    if (
        order_data.time_in_force not in BybitExchange.futures_time_in_forces.keys()
        and order_data.time_in_force
        not in BybitExchange.futures_time_in_forces.values()
    ):
        raise ValueError("'timeInForce' is not correct.")


def check_futures_order_type(order_data):
    if order_data.order_type is None:
        raise ValueError("Specify order type.")

    if order_data.order_type not in BybitExchange.futures_order_types:
        raise ValueError(
            "Bad order type specified. Available order types for futures: {}".format(
                BybitExchange.futures_order_types
            )
        )


def check_futures_order_side(order_data):
    if order_data.side is None:
        raise ValueError("Specify order side.")


def check_futures_order_quantity(order_data):
    if order_data.quantity is None:
        raise ValueError("Specify quantity.")


def check_futures_order_symbol(order_data):
    if order_data.symbol is None:
        raise ValueError("Specify symbol.")


class PyBitHTTP(HTTP):
    def __init__(
        self,
        testnet=False,
        api_key=None,
        api_secret=None,
    ):
        super().__init__(
            testnet=testnet,
            api_key=api_key,
            api_secret=api_secret,
        )

    def query_history_order(self, **kwargs):
        if self.spot is True:
            suffix = "/spot/v1/history-orders"

            return self._submit_request(
                method="GET", path=self.endpoint + suffix, query=kwargs, auth=True
            )
        else:
            raise NotImplementedError("Not implemented for futures market.")

    def get_active_order_spot(self, **kwargs):
        if self.spot is True:
            suffix = "/spot/v1/order"

            return self._submit_request(
                method="GET", path=self.endpoint + suffix, query=kwargs, auth=True
            )
        else:
            raise NotImplementedError("Not implemented for futures market.")


class BybitExchange(BaseExchange):
    time_indexes_in_candle_data = [0, 6]
    desired_candle_data_indexes = [0, 1, 2, 3, 4, 5, 6, 8]

    spot_order_types = ["LIMIT", "MARKET", "LIMIT_MAKER"]
    spot_time_in_forces = ["GTC", "FOK", "IOC"]

    futures_order_types = ["MARKET", "LIMIT", "STOP_MARKET", "STOP_LIMIT"]
    futures_time_in_forces = {
        "GTC": "GoodTillCancel",
        "IOC": "ImmediateOrCancel",
        "FIK": "FillOrKill",
        "PO": "PostOnly",
    }

    time_intervals = [
        "1m",
        "3m",
        "5m",
        "15m",
        "30m",
        "1h",
        "2h",
        "4h",
        "6h",
        "12h",
        "1d",
        "1w",
        "1M",
    ]

    def __init__(self, credentials, sandbox=False, unified_in_outs=True):
        self.api_key = credentials["spot"]["key"]
        self.secret = credentials["spot"]["secret"]
        self.sandbox = sandbox
        self.unified_in_outs = unified_in_outs

        if sandbox:
            self.spot_session = PyBitHTTP(
                testnet=True, api_key=self.api_key, api_secret=self.secret
            )
            self.futures_session = PyBitHTTP(
                testnet=True,
                api_key=self.api_key,
                api_secret=self.secret,
            )
        else:
            self.spot_session = PyBitHTTP(
                testnet=False, api_key=self.api_key, api_secret=self.secret
            )
            self.futures_session = PyBitHTTP(
                testnet=False, api_key=self.api_key, api_secret=self.secret
            )

    def get_balance(self, asset="", futures=False):
        if futures:
            return (
                BybitHelpers.get_balance_out(
                    self.futures_session.get_wallet_balance(accountType="CONTRACT")[
                        "result"
                    ],
                    futures=True,
                )
                if asset in [None, ""]
                else BybitHelpers.get_balance_out(
                    self.futures_session.get_wallet_balance(
                        accountType="CONTRACT", coin=asset
                    )["result"],
                    single=True,
                    futures=True,
                )
            )
        if asset in [None, ""]:
            return BybitHelpers.get_balance_out(
                self.spot_session.get_wallet_balance(accountType="SPOT")["result"][
                    "balances"
                ]
            )
        assets = self.spot_session.get_wallet_balance(accountType="SPOT")["result"][
            "balances"
        ]
        for coin in assets:
            if asset == coin["coin"]:
                return BybitHelpers.get_balance_out(coin, single=True)

        try:
            self.futures_session.get_wallet_balance(coin=asset)
            return BybitHelpers.make_dummy_balance(asset)
        except Exception as e:
            raise ValueError("Coin not found.") from e

    def symbol_account_trade_history(
        self, symbol, futures=False, from_id=None, limit=None
    ):
        if futures:
            trade_history = self.futures_session.user_trade_records(
                symbol=symbol, limit=limit, fromId=from_id
            )
            return BybitHelpers.get_my_trade_history_out(
                trade_history["result"]["data"], futures=True
            )
        else:
            trade_history = self.spot_session.user_trade_records(
                symbol=symbol, limit=limit, fromId=from_id
            )
            return BybitHelpers.get_my_trade_history_out(trade_history["result"])

    def test_spot_order(self, order_data: DataHelpers.OrderData):
        is_spot_order_data_valid(order_data)

        # sourcery skip: no-conditionals-in-tests
        if (
            order_data.iceberg_qty is not None
            or order_data.new_order_resp_type is not None
            or order_data.quote_order_qty is not None
            or order_data.recv_window is not None
            or order_data.stop_price is not None
        ):
            warnings.warn("Some of the given parameters have no use in ByBit exchange.")

        return order_data

    def make_spot_order(self, order_data):
        order_params = BybitHelpers.get_spot_order_as_dict(order_data)

        return BybitHelpers.get_make_spot_order_out(
            self.spot_session.place_active_order(**order_params)["result"]
        )

    def create_and_test_spot_order(
        self,
        symbol,
        side,
        order_type,
        quantity=None,
        price=None,
        time_in_force=None,
        stop_price=None,
        iceberg_qty=None,
        new_order_resp_type=None,
        new_client_order_id=None,
        extra_params=None,
    ):
        curr_order = DataHelpers.set_spot_order_data(
            iceberg_qty,
            new_client_order_id,
            new_order_resp_type,
            order_type,
            price,
            quantity,
            side,
            stop_price,
            symbol,
            time_in_force,
            extra_params,
        )

        self.test_spot_order(curr_order)

        return curr_order

    def get_symbol_orders(
        self,
        symbol,
        futures=False,
        order_id=None,
        start_time=None,
        end_time=None,
        limit=None,
    ):
        if futures:
            history_list = []
            page_number = 1
            end_time_string = None
            start_time_string = None
            done = False

            while not done:
                history = self.futures_session.get_active_order(
                    symbol=symbol, page=page_number, limit=50
                )

                if start_time is not None:
                    start_time_string = start_time.strftime("%Y-%m-%dT%H:%M:%SZ")
                if end_time is not None:
                    end_time_string = end_time.strftime("%Y-%m-%dT%H:%M:%SZ")

                for order in history["result"]["data"]:
                    if end_time is not None and end_time_string < order["create_time"]:
                        continue

                    if (
                        start_time is not None
                        and order["created_time"] < start_time_string
                    ):
                        done = True
                        break

                    history_list.append(order)

                if limit is not None and limit <= len(history_list):
                    done = True

                if len(history["result"]["data"]) < 50:
                    done = True

                page_number += 1

            return history_list
        else:
            history = self.spot_session.query_history_order(
                symbol=symbol,
                orderId=order_id,
                startTime=start_time,
                endtime=end_time,
                limit=limit,
            )
            return history["result"]

    def get_open_orders(self, symbol, futures=False):
        if futures:
            return self.get_futures_open_orders(symbol)
        else:
            return self.get_spot_open_orders(symbol)

    def get_spot_open_orders(self, symbol):
        open_orders = (
            self.spot_session.query_active_order()["result"]
            if symbol is None
            else self.spot_session.query_active_order(symbol=symbol)["result"]
        )
        return BybitHelpers.get_open_orders_out(open_orders)

    def get_futures_open_orders(self, symbol):
        open_active_orders = self.futures_session.query_active_order(symbol=symbol)
        open_orders = [
            BybitHelpers.futures_order_out(activeOrder)
            for activeOrder in open_active_orders["result"]
        ]
        open_conditional_orders = self.futures_session.query_conditional_order(
            symbol=symbol
        )
        open_orders.extend(
            BybitHelpers.futures_order_out(conditionalOrder, is_conditional=True)
            for conditionalOrder in open_conditional_orders["result"]
        )
        return open_orders

    def cancel_all_symbol_open_orders(self, symbol, futures=False):
        if futures:
            result = self.futures_session.cancel_all_active_orders(symbol=symbol)
            canceled_orders_ids = [result["result"]]
            result = self.futures_session.cancel_all_conditional_orders(symbol=symbol)
            canceled_orders_ids.append(result["result"])
        else:
            result = self.spot_session.batch_fast_cancel_active_order(
                symbol=symbol, orderTypes="LIMIT,LIMIT_MAKER,MARKET"
            )
            return result["result"]["success"]

    def cancel_order(self, symbol, order_id=None, local_order_id=None, futures=False):
        if futures:
            return self.cancel_futures_order(local_order_id, order_id, symbol)
        else:
            return self.cancel_spot_order(local_order_id, order_id)

    def cancel_spot_order(self, local_order_id, order_id):
        if order_id is not None:
            result = self.spot_session.cancel_active_order(orderId=order_id)
        elif local_order_id is not None:
            result = self.spot_session.cancel_active_order(orderLinkId=local_order_id)
        else:
            raise ValueError("Must specify either 'orderId' or 'localOrderId'")
        return result

    def cancel_futures_order(self, local_order_id, order_id, symbol):
        if order_id is not None:
            try:
                result = self.futures_session.cancel_active_order(
                    symbol=symbol, order_id=order_id
                )
            except Exception as e:
                try:
                    result = self.futures_session.cancel_conditional_order(
                        symbol=symbol, order_id=order_id
                    )
                except Exception as e:
                    raise RuntimeError(
                        "Problem in canceling order in bybit: {}".format(str(e))
                    ) from e
        elif local_order_id is not None:
            try:
                result = self.futures_session.cancel_active_order(
                    symbol=symbol, order_link_id=local_order_id
                )
            except Exception as e:
                try:
                    result = self.futures_session.cancel_conditional_order(
                        symbol=symbol, order_link_id=local_order_id
                    )
                except Exception as e:
                    raise RuntimeError(
                        "Problem in canceling order in bybit: {}".format(str(e))
                    ) from e
        else:
            raise ValueError("Must specify either 'orderId' or 'localOrderId'")
        return result

    def get_order(self, symbol, order_id=None, local_order_id=None, futures=False):
        if futures:
            order = self.get_futures_order(local_order_id, order_id, symbol)
            return BybitHelpers.futures_order_out(order["result"])
        else:
            order = self.get_spot_order(local_order_id, order_id)
            return BybitHelpers.get_order_out(order)

    def get_spot_order(self, local_order_id, order_id):
        if order_id is not None:
            try:
                order = self.spot_session.get_active_order_spot(orderId=order_id)[
                    "result"
                ]
            except Exception as e:
                raise RuntimeError("Problem in fetching order from bybit.") from e
        elif local_order_id is not None:
            try:
                order = self.spot_session.get_active_order_spot(
                    orderLinkId=local_order_id
                )["result"]
            except Exception as e:
                raise RuntimeError("Problem in fetching order from bybit.") from e
        else:
            raise ValueError(
                "Specify either order Id in the exchange or local Id sent with the order"
            )
        return order

    def get_futures_order(self, local_order_id, order_id, symbol):
        if order_id is not None:
            try:
                order = self.futures_session.query_active_order(
                    symbol=symbol, order_id=order_id
                )
            except Exception as e:
                try:
                    order = self.futures_session.query_conditional_order(
                        symbol=symbol, order_id=order_id
                    )
                except Exception as e:
                    raise RuntimeError(
                        "Problem in fetching order from bybit: {}".format(str(e))
                    ) from e
        elif local_order_id is not None:
            try:
                order = self.futures_session.query_active_order(
                    symbol=symbol, order_link_id=local_order_id
                )
            except Exception as e:
                try:
                    order = self.futures_session.query_conditional_order(
                        symbol=symbol, order_link_id=local_order_id
                    )
                except Exception as e:
                    raise RuntimeError(
                        "Problem in fetching order from bybit: {}".format(str(e))
                    ) from e
        else:
            raise ValueError(
                "Specify either order Id in the exchange or local Id sent with the order"
            )
        return order

    def get_trading_fees(self, futures=False):
        raise NotImplementedError()

    def get_symbol_ticker_price(self, symbol, futures=False):
        if futures:
            symbol_info = self.futures_session.latest_information_for_symbol(
                symbol=symbol
            )["result"]
            return float(symbol_info[0]["last_price"])
        else:
            symbol_info = self.spot_session.latest_information_for_symbol(symbol=symbol)
            return float(symbol_info["result"]["lastPrice"])

    def get_symbol_klines(
        self,
        symbol,
        interval,
        start_time=None,
        end_time=None,
        limit=None,
        futures=False,
        blvtnav=False,
        convert_date_time=False,
        do_clean=False,
        to_clean_dataframe=False,
    ):
        if interval not in self.time_intervals:
            raise ValueError("Time interval is not valid.")

        if futures:
            data = self._get_futures_symbol_klines(interval, limit, start_time, symbol)
        else:
            data = self._get_spot_symbol_klines(
                end_time, interval, limit, start_time, symbol
            )

        if convert_date_time or to_clean_dataframe:
            BybitHelpers.klines_convert_date(
                data, futures, self.time_indexes_in_candle_data
            )

        if not do_clean and not to_clean_dataframe:
            return data
        final_data_array = BybitHelpers.get_klines_desired_only_cols(
            data, self.desired_candle_data_indexes
        )

        if to_clean_dataframe:
            return BybitHelpers.klines_convert_to_pandas(final_data_array)
        return final_data_array

    def _get_spot_symbol_klines(self, end_time, interval, limit, start_time, symbol):
        start_timestamp = None if start_time is None else start_time.timestamp() * 1000
        end_timestamp = end_time.timestamp() * 1000 if end_time is not None else None

        if limit is not None:
            if limit > 1000:
                limit = 1000
            elif limit < 1:
                limit = 1

        data = self.spot_session.query_kline(
            symbol=symbol,
            interval=interval,
            startTime=start_timestamp,
            endTime=end_timestamp,
            limit=limit,
        )["result"]
        for datum in data:
            for idx in range(len(datum)):
                if idx in self.time_indexes_in_candle_data:
                    continue
                datum[idx] = float(datum[idx])
        return data

    def _get_futures_symbol_klines(self, interval, limit, start_time, symbol):
        futures_interval = BybitHelpers.convert_interval_to_futures_klines(interval)
        data = []

        if limit is not None and limit > 200 or limit is None:
            limit = 200
        elif limit < 1:
            limit = 1

        if start_time is None:
            start_timestamp = int(
                datetime.now().timestamp()
                - BybitHelpers.get_interval_in_seconds(interval, self.time_intervals)
                * limit
            )
        elif isinstance(start_time, int):
            start_timestamp = start_time
        else:
            start_timestamp = int(start_time.timestamp)

        candles = self.futures_session.query_kline(
            symbol=symbol,
            interval=futures_interval,
            from_time=start_timestamp,
            limit=limit,
        )
        for candle in candles["result"]:
            data_array = [
                float(candle["open_time"]),
                float(candle["open"]),
                float(candle["high"]),
                float(candle["low"]),
                float(candle["close"]),
                float(candle["volume"]),
                int(candle["open_time"])
                + BybitHelpers.get_interval_in_seconds(interval, self.time_intervals),
                None,
                None,
                None,
                None,
            ]
            data.append(data_array)
        return data

    def get_exchange_time(self, futures=False):
        if futures:
            return self.futures_session.server_time()["time_now"]
        else:
            return int(self.spot_session.server_time()["result"]["serverTime"])

    def get_symbol_24h_ticker(self, symbol):
        raise NotImplementedError()

    def test_futures_order(self, futures_order_data):
        # sourcery skip: no-conditionals-in-tests
        if futures_order_data.time_in_force is None:
            futures_order_data.time_in_force = "GoodTillCancel"

        if futures_order_data.close_position is None:
            futures_order_data.close_position = False

        if futures_order_data.reduce_only is None:
            futures_order_data.reduce_only = False

        is_futures_order_data_valid(futures_order_data)

        return futures_order_data

    def make_futures_order(self, futures_order_data: DataHelpers.FuturesOrderData):
        order_params = BybitHelpers.get_futures_order_as_dict(
            futures_order_data, self.futures_time_in_forces
        )

        if "STOP" in futures_order_data.orderType:
            result = self.futures_session.place_conditional_order(**order_params)
            return BybitHelpers.futures_order_out(result["result"], is_conditional=True)
        else:
            result = self.futures_session.place_active_order(**order_params)
            return BybitHelpers.futures_order_out(result["result"])

    def create_and_test_futures_order(
        self,
        symbol,
        side,
        order_type,
        position_side=None,
        time_in_force=None,
        quantity=None,
        reduce_only=None,
        price=None,
        new_client_order_id=None,
        stop_price=None,
        close_position=None,
        activation_price=None,
        callback_rate=None,
        working_type=None,
        price_protect=None,
        new_order_resp_type=None,
        extra_params=None,
        quote_quantity=None,
    ):
        curr_order = DataHelpers.set_futures_order_data(
            activation_price,
            callback_rate,
            close_position,
            extra_params,
            new_client_order_id,
            new_order_resp_type,
            order_type,
            position_side,
            price,
            price_protect,
            quantity,
            reduce_only,
            side,
            stop_price,
            symbol,
            time_in_force,
            working_type,
            quote_quantity,
        )

        self.test_futures_order(curr_order)

        return curr_order

    def make_batch_futures_order(self, futures_order_datas):
        batch_orders = []
        batch_conditional_orders = []
        for order in futures_order_datas:
            order_as_dict = BybitHelpers.get_futures_order_as_dict(
                order, self.futures_time_in_forces
            )

            if "STOP" in order.order_type:
                batch_conditional_orders.append(order)
            else:
                batch_orders.append(order_as_dict)

        results = []
        if batch_conditional_orders:
            put_results = self.futures_session.place_conditional_order_bulk(
                batch_conditional_orders
            )
            results.extend(
                BybitHelpers.futures_order_out(result["result"], is_conditional=True)
                for result in put_results
            )
        if batch_orders:
            put_results = self.futures_session.place_active_order_bulk(batch_orders)
            results.extend(
                BybitHelpers.futures_order_out(result["result"])
                for result in put_results
            )
        return results

    def change_initial_leverage(self, symbol, leverage):
        return self.futures_session.set_leverage(
            symbol=symbol, buy_leverage=leverage, sell_leverage=leverage
        )["result"]

    def change_margin_type(self, symbol, margin_type, params):
        try:
            buy_leverage = params["buyLeverage"]
            sell_leverage = params["sellLeverage"]
            if margin_type.upper() == "ISOLATED":
                is_isolated = True
            elif margin_type.upper() == "CROSS":
                is_isolated = False
            else:
                raise ValueError("Margin type must either be 'ISOLATED' or 'CROSS'.")
        except Exception as e:
            raise ValueError(
                "Must specify 'buyLeverage' and 'sellLeverage' in 'params"
            ) from e

        self.futures_session.cross_isolated_margin_switch(
            symbol=symbol,
            is_isolated=is_isolated,
            buy_leverage=buy_leverage,
            sell_leverage=sell_leverage,
        )
        return True

    def change_position_margin(self, symbol, amount):
        result = self.futures_session.change_margin(symbol=symbol, margin=amount)
        return result["ret_code"] == 0

    def get_position(self):
        return self.futures_session.my_position()["result"]

    def spot_best_bid_asks(self, symbol=None):
        return self.spot_session.best_bid_ask_price(symbol=symbol)["result"]

    def get_symbol_order_book(self, symbol, limit=None, futures=False):
        if futures:
            return self.futures_session.orderbook(symbol=symbol)["result"]
        else:
            return self.spot_session.orderbook(symbol=symbol)["result"]

    def get_symbol_recent_trades(self, symbol, limit=None, futures=False):
        if futures:
            limit = min(limit, 1000) if limit is not None and limit > 0 else 500

            recent_trades = self.futures_session.public_trading_records(
                symbol=symbol, limit=limit
            )["result"]
            return BybitHelpers.get_recent_trade_history_out(
                recent_trades, futures=True
            )
        else:
            limit = min(limit, 60) if limit is not None and limit > 0 else 60

            recent_trades = self.spot_session.public_trading_records(
                symbol=symbol, limit=limit
            )["result"]
            return BybitHelpers.get_recent_trade_history_out(recent_trades)

    def get_position_info(self, symbol=None):
        result = self.futures_session.my_position(symbol=symbol)
        return result["result"]

    def get_symbol_min_trade(self, symbol, futures=False):
        symbol_ticker_price = self.get_symbol_ticker_price(
            symbol=symbol, futures=futures
        )

        min_quantity = None
        min_quote_quantity = None
        step_quantity = None
        step_price = None

        if futures:
            symbol_infos = self.futures_session.query_symbol()["result"]

            for symbolInfo in symbol_infos:
                if symbolInfo["name"] == symbol:
                    min_quantity = float(
                        symbolInfo["lot_size_filter"]["min_trading_qty"]
                    )
                    min_quote_quantity = symbol_ticker_price * min_quantity
                    step_quantity = float(symbolInfo["lot_size_filter"]["qty_step"])
                    step_price = symbolInfo["price_filter"]["tick_size"]

        else:
            symbol_infos = self.spot_session.query_symbol()["result"]

            for symbolInfo in symbol_infos:
                if symbolInfo["name"] == symbol:
                    min_quantity = float(symbolInfo["minTradeQuantity"])
                    min_quote_quantity = float(symbolInfo["minTradeAmount"])
                    step_quantity = float(symbolInfo["basePrecision"])
                    step_price = symbolInfo["minPricePrecision"]

        return {
            "minQuantity": min_quantity,
            "minQuoteQuantity": min_quote_quantity,
            "precisionStep": step_quantity,
            "stepPrice": step_price,
        }

    def make_sl_tp_limit_futures_order(
        self,
        symbol,
        order_side,
        quantity=None,
        quote_quantity=None,
        enter_price=None,
        take_profit=None,
        stop_loss=None,
        leverage=None,
        margin_type=None,
    ):
        if (
            enter_price is None
            or take_profit is None
            or stop_loss is None
            or leverage is None
            or margin_type is None
        ):
            raise ValueError("Specify all inputs.")

        symbol_info = self.get_symbol_min_trade(symbol=symbol, futures=True)

        quantity = DataHelpers.get_quantity(
            enter_price, quantity, quote_quantity, symbol_info["precisionStep"]
        )

        self.set_margin_leverage(leverage, margin_type, symbol)
        order_params = {
            "takeProfit": take_profit,
            "stopLoss": stop_loss,
            "tpTriggerBy": "MarkPrice",
            "slTriggerBy": "MarkPrice",
        }

        futures_order = self.create_and_test_futures_order(
            symbol=symbol,
            side=order_side,
            order_type="LIMIT",
            time_in_force="GTC",
            quantity=quantity,
            price=enter_price,
            close_position=False,
            extra_params=order_params,
        )
        ordering_result = self.make_futures_order(futures_order)
        return ordering_result["orderId"]

    def set_margin_leverage(self, leverage, margin_type, symbol):
        params = {"buyLeverage": leverage, "sellLeverage": leverage}
        self.change_initial_leverage(leverage, symbol)
        self.change_margin_type(symbol, margin_type, params)

    def get_symbol_24h_changes(self, futures=False):
        changes_list = []
        if futures:
            for ticker in self.futures_session.latest_information_for_symbol()[
                "result"
            ]:
                print(ticker)
                if not ticker["is_block_trade"]:
                    changes_list.append((ticker["symbol"], 0))
        else:
            symbol_infos = self.spot_session.query_symbol()["result"]
            changes_list.extend((ticker["name"], 0) for ticker in symbol_infos)
        return sorted(changes_list, key=lambda x: x[1], reverse=True)

    def get_symbol_list(self, futures=False):
        raise NotImplementedError()

    def get_latest_symbol_names(self, num_of_symbols=None, futures=False):
        raise NotImplementedError()

    def get_income_history(
        self, symbol, income_type=None, start_time=None, end_time=None, limit=None
    ):
        args = {
            "start_date": start_time,
            "end_date": end_time,
            "wallet_fund_type": income_type,
            "limit": limit,
            "coin": symbol,
        }
        args = {k: v for k, v in args.items() if v is not None}

        return self.futures_session.wallet_fund_records(**args)

    def get_long_short_ratios(
        self, symbol, period, limit=None, start_time=None, end_time=None
    ):
        raise NotImplementedError()

    def get_deposit_address(self, coin, network=None):
        raise NotImplementedError()

    def withdraw(self, coin, address, amount, extra_data):
        raise NotImplementedError()
